/**************************************************************************

    Ear Trainer - A simple Ear training tool for the Palm Computing Platform
    Copyright (C) 1999 Sam Trenholme

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

**************************************************************************/

#include <Pilot.h>				
#include "Rachel.h"
#include "RachelRsc.h"
#ifdef __GNUC__
#include "Callbacks.h"
#endif		

// for definition of IsDigit
#include <CharAttr.h>
#include <SysEvtMgr.h>

// Global variables
// Variables that start with 'RachelPref' are preferences stored between
// sessions.  'RachelGlobal' variables are global variables not saved
// between sessions.
// RachelPref1stNote:   The first note Rachel played last time
// RachelPrefInterval:  The interval Rachel played last time
// RachelPrefGuess:     The last interval tried by the user
// RachelPrefVolume:    The volume Rachel sings at
// RachelPrefDuration:  Whether to sing an eight, quarter, or 1/2 note
// RachelPrefRndDrtn:   Save random duration for "play interval again"
// RachelPrefDirection: Whether to make the sequence acending or decending
// RachelPrefRndDir:    Save random direction for "play interval again"
// RachelPrefAlertOn:   Do we alert on Wrong interval, right interval, or both?
int RachelPref1stNote, RachelPrefInterval, RachelPrefGuess, RachelPrefVolume,
    RachelPrefDuration, RachelPrefRndDrtn, RachelPrefDirection,
    RachelPrefRndDir, RachelPrefAlertOn;

int RachelGlobalGuessed;

// This function is given an ID of a list (RachelIntervalList, etc),
// a list selection number, and optionally a pointer to a form.  
// It returns a pointer to the string for the list selection in question
char *RachelListText(Word list_id, Word list_item, FormPtr formP)
{
	FormPtr frm = NULL;
	ListPtr list = NULL;

	// If they did not supply a form pointer (gave it a NULL value),
        // then have it point to the currently active form
	if(formP == NULL)
		frm = FrmGetActiveForm();
	else
		frm = formP;

	// Just in case we don't have an active form, bail to stop a
        // fatal exception
	if(frm == NULL)
		return NULL;

	// Get a pointer to the list
	if(frm != NULL)
		list = FrmGetObjectPtr(
		       frm, FrmGetObjectIndex(frm, list_id));

	if(list != NULL)
		return LstGetSelectionText(list, list_item);

	// If there was invalid data, then return a pointer to NULL
	return NULL;

}

// This function changes the text label for the selected value in a popup field
// Input: text to put in popup field ; Object Index for popup ;
// pointer to form (if NULL then use currently active form)
void RachelChangePopupText(char *newtext, Word popup_object, FormPtr frm)
{

	ControlPtr ctrl;

	if(frm == NULL)
		frm = FrmGetActiveForm();

	// If it is still NULL (no active form), return to stop fatal exception
	if(frm == NULL)
		return;

        ctrl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, popup_object));

	if(ctrl != NULL && newtext != NULL)
		CtlSetLabel(ctrl,newtext);
	
}

// This function changes the alert text for the currently active form
// Input: pointer to currently active form
void RachelChangeAlertPopup(FormPtr frm)
{

	if(frm == NULL)
		frm = FrmGetActiveForm();

	// If frm is still NULL, return to stop fatal exception
	if(frm == NULL)
		return;
	
	if(RachelPrefAlertOn == RACHEL_ALERTON_RIGHT)
		RachelChangePopupText(RACHEL_ALERTON_MSG_RIGHT,
				      RachelAlertTrigger, frm);
	else if(RachelPrefAlertOn == RACHEL_ALERTON_WRONG)
		RachelChangePopupText(RACHEL_ALERTON_MSG_WRONG,
				      RachelAlertTrigger, frm);
        else if(RachelPrefAlertOn == RACHEL_ALERTON_BOTH)
		RachelChangePopupText(RACHEL_ALERTON_MSG_BOTH,
				      RachelAlertTrigger, frm);
	else if(RachelPrefAlertOn == RACHEL_ALERTON_NEITHER)
		RachelChangePopupText(RACHEL_ALERTON_MSG_NEITHER,
				      RachelAlertTrigger, frm);
	else // This should never happen, but if it does
		RachelChangePopupText(" ",RachelAlertTrigger, frm);

}

// This function restores the values of fields on the currently active form
// to their values in the preferences.  This is used when we start running 
// the application
void RachelRestorePrefs()
{

	// Some variables used for setting the element on the list
	FormPtr frm;
	ListPtr list;
	ControlPtr ctrl;
	
	// Point to the currently aactive form
	frm = FrmGetActiveForm();

	// Point to the interval selection list
	if(frm != NULL)
		list = FrmGetObjectPtr(
                       frm, FrmGetObjectIndex(frm, RachelIntervalList));
	else
		return; // We exit now to minimize code size in case frm
                        // was not defined

	// Restore the previously selected interval from the prefs
	if(list != NULL)
		LstSetSelection(list,RachelPrefGuess);

	// Restore the previously selected volume
	if(RachelPrefVolume <= 0) // This should never happen, but if it does
		RachelChangePopupText(" ",RachelVolumeTrigger,frm);
	else if(RachelPrefVolume <= RACHEL_VOLUME_LOW)
		RachelChangePopupText(RachelListText(RachelVolumePopup,0,frm),
                                      RachelVolumeTrigger,frm);
	else if(RachelPrefVolume <= RACHEL_VOLUME_MID)
		RachelChangePopupText(RachelListText(RachelVolumePopup,1,frm),
				      RachelVolumeTrigger,frm);
	else
		RachelChangePopupText(RachelListText(RachelVolumePopup,2,frm),
				      RachelVolumeTrigger,frm);

	// Restore the previously selected duration	
	if(RachelPrefDuration >= RACHEL_DURATION_SHORT && 
           RachelPrefDuration <= RACHEL_DURATION_MAX)
		RachelChangePopupText(RachelListText(RachelDurationPopup,
						     RachelPrefDuration,
						     frm), 
                                      RachelDurationTrigger,frm);

	// Restore the previously selected direction
	if(RachelPrefDirection >= RACHEL_DIRECTION_UP &&
	   RachelPrefDirection <= RACHEL_DIRECTION_MAX)
		RachelChangePopupText(RachelListText(RachelDirectionPopup,
				              	     RachelPrefDirection,
						     frm),
                                      RachelDirectionTrigger,frm);

	// Restore the previously selected "alert on" preference
	RachelChangeAlertPopup(frm);

                       
}

// Restore all the RachelPref variables to their values stored in the 
// preference database
static Err RachelStartApplication(void)
{

	// Read Interval and RachelPref1stNote from global preferences
	int ReadPref[RACHEL_PREFERENCES_SAVED + 1];
	Word size = (RACHEL_PREFERENCES_SAVED  + 1) * sizeof(int);

	ReadPref[RACHEL_PREFERENCES_SAVED] = 0;

	PrefGetAppPreferences('Rchl',0,ReadPref,&size,true);

	if(size != (RACHEL_PREFERENCES_SAVED + 1) * sizeof(int) || 
           ReadPref[RACHEL_PREFERENCES_SAVED] != 12)
		{
		// Initialize the global variables if we didn't get prefs
		RachelPrefInterval = RachelPref1stNote = -1;
		RachelPrefGuess = 0;
		RachelPrefVolume = RACHEL_VOLUME_HIGH;
		RachelPrefDuration = RACHEL_DURATION_MED;
		RachelPrefRndDrtn = RACHEL_DURATION_MED;
		RachelPrefDirection = RACHEL_DIRECTION_UP;
		RachelPrefRndDir = RACHEL_DIRECTION_UP;
		RachelPrefAlertOn = RACHEL_ALERTON_BOTH;
		}
	else
		{
		// Otherwise, these global variables are obtained from the
		// prefs
		RachelPref1stNote   = ReadPref[0];
		RachelPrefInterval  = ReadPref[1];
		RachelPrefGuess     = ReadPref[2];
		RachelPrefVolume    = ReadPref[3];
		RachelPrefDuration  = ReadPref[4];
		RachelPrefRndDrtn   = ReadPref[5];
		RachelPrefDirection = ReadPref[6];
		RachelPrefRndDir    = ReadPref[7];	
		RachelPrefAlertOn   = ReadPref[8];
		}

	// Set the value of the globa variables not saved between sessions
	RachelGlobalGuessed = true;

	// Open up the main form
	FrmGotoForm(RachelForm);

	return 0;
}


// Store current Interval to guess when leaving the application
static void RachelStopApplication(void)
{
	// Save the Interval and RachelPref1stNote

	int SavePref[RACHEL_PREFERENCES_SAVED + 1];

	// We have a marker that all the preferences for this version have 
        // been saved.  When upgrading to future versions, this guarantees
        // that the preferences are reset between upgrades
	SavePref[RACHEL_PREFERENCES_SAVED] = 12;

	SavePref[0] = RachelPref1stNote;
	SavePref[1] = RachelPrefInterval;
	SavePref[2] = RachelPrefGuess;
	SavePref[3] = RachelPrefVolume;
	SavePref[4] = RachelPrefDuration;
	SavePref[5] = RachelPrefRndDrtn;
	SavePref[6] = RachelPrefDirection;
	SavePref[7] = RachelPrefRndDir;
	SavePref[8] = RachelPrefAlertOn;

	PrefSetAppPreferences('Rchl',0,1,SavePref,
                              (RACHEL_PREFERENCES_SAVED + 1) * sizeof(Int),
                              true);

	FrmCloseAllForms(); 

}

// This converts the duration stored in the preferences to milliseconds
UInt RachelConvertDuration(int pref_duration)
{

	// If it is a "random" duration, get the random duration from the
        // prefs.  We do it this way because we make RachelPrefRndDrtn random
        // every time we play a new interval, and want to repeat that random
        // duration every time we hit "Play interval again"
	if(pref_duration == RACHEL_DURATION_RANDOM)
		pref_duration = RachelPrefRndDrtn;

	// The values in qustion are an eight note, a quarter note, and a
        // half note at 120 BPM
	switch(pref_duration)
	{
	case RACHEL_DURATION_SHORT:
		return 250;
	case RACHEL_DURATION_MED:
		return 500;
	case RACHEL_DURATION_LONG:
		return 1000;
	}

	// Just in case the value is completely hozed
	return 800;

}

// Given a note pair, return that note pair in the direciton specified
// in the direciton preference.  
RachelNotePair RachelNotesInDirection(RachelNotePair notes)
{

	int tempNote; // Tempory variable so we can swap if needed

	// If the direction should be up
	if(RachelPrefDirection == RACHEL_DIRECTION_UP ||
           (RachelPrefDirection == RACHEL_DIRECTION_RANDOM &&
            RachelPrefRndDir == RACHEL_DIRECTION_UP))
		{
		// And the direction is actually down
		if(notes.note1 > notes.note2)
			{
			// Swap the notes
			tempNote = notes.note1;
			notes.note1 = notes.note2;
			notes.note2 = tempNote;
			}
		}
	else // Otherwise, the direction is going down
		if(notes.note1 < notes.note2)
			{
			// Swap the notes
			tempNote = notes.note1;
			notes.note1 = notes.note2;
			notes.note2 = tempNote;
			}

	return notes;
			
}

// Given a first note and an interval, Return two absolute notes within
// Rachel's singing range
RachelNotePair RachelIntervalInRange(int local_note, int local_interval)
{

	int secondNote;
	RachelNotePair ret;

	ret.note1 = ret.note2 = -1;

	// Normally, we play first_note, then first_note + interval, however
	// if the note + interval puts us out of bounds
	if(local_note + local_interval > RACHEL_HIGH_NOTE ||
           local_note + local_interval < RACHEL_LOW_NOTE) // Interval could
                                                          // be negative
		{
		// If subtracting the note by the interval also puts
                // us out of bounds, return invalid notes
		if(local_note - local_interval < RACHEL_LOW_NOTE ||
                   local_note - local_interval > RACHEL_HIGH_NOTE)
			return ret;
		// Otherwise, make it so the first note is the interval 
                // below the selected note.
		secondNote = local_note;
		local_note -= local_interval;
		}
	// If note + interval is in bounds, we return note and note + interval
	else
		secondNote = local_interval + local_note;

	// Set the value of the notes in question
	ret.note1 = local_note;
	ret.note2 = secondNote;

	// Make sure the notes are ordered in the user-specified direction
	ret = RachelNotesInDirection(ret);

	return ret;

}

// Given a certain MIDI note number, and a pointer to a string, make the
// string have the given note.  If it is sharp/flat, name it as a sharp
char *RachelNameNoteSharp(int note, char *string)
{

	// Just in case we point to NULL, this may save us from a 
        // fatal exception
	if(string == NULL)
		return NULL;

	*(string + 1) = '\0';

	switch (note % 12)
	{
		case 1:
			*(string + 1) = '#';
		case 0:
			*string = 'C';
			break;
		case 3:
			*(string + 1) = '#';
		case 2:
			*string = 'D';
			break;
		case 4:
			*string = 'E';
			break;
		case 6:
			*(string + 1) = '#';
		case 5:
			*string = 'F';
			break;
		case 8:
			*(string + 1) = '#';
		case 7:
			*string = 'G';
			break;
		case 10:
			*(string + 1) = '#';
		case 9:
			*string = 'A';
			break;
		case 11:
			*string = 'B';
			break;
	}

	if(*(string + 1) == '#');
		*(string + 2) = '\0';

	return string;

}

// Given two notes, show an alert informing the user what the two notes are
void RachelNameNotes(int note1, int note2)
{

	char n1[3],n2[3];

	// If both notes are within bounds
	if(note1 >= RACHEL_LOW_NOTE && note1 <= RACHEL_HIGH_NOTE &&
	   note2 >= RACHEL_LOW_NOTE && note2 <= RACHEL_HIGH_NOTE)
	  	 
		FrmCustomAlert(RachelNameNotesPlayed,
                               RachelNameNoteSharp(note1,n1),
                               RachelNameNoteSharp(note2,n2), NULL);
	// Otherwise tell them to have Rachel sing a random interval first
	else
		FrmAlert(RachelNoNotesToName);

}

// Given a note and a interval, name the notes that Rachel will actually sing 
void RachelNameNotesInInterval(int note, int interval)
{

	RachelNotePair notes;
	notes = RachelIntervalInRange(note,interval);
	RachelNameNotes(notes.note1,notes.note2);

}

// Generate a random note in Rachel's singing range
int RachelRandomNote()
{

	return (SysRandom(0) % (RACHEL_HIGH_NOTE - RACHEL_LOW_NOTE)) 
               + RACHEL_LOW_NOTE;

}

// Generate a random interval to sing
int RachelRandomInterval()
{

	// This returns an interval between a semitone and an octave
	return (SysRandom(0) % 12) + 1;

}

// Generate a random duration to play
int RachelRandomDuration()
{

	return SysRandom(0) % RACHEL_DURATION_MAX;

}

// Generate a random direction to play in
int RachelRandomDirection()
{

	return SysRandom(0) % RACHEL_DIRECTION_MAX;

}

// Determine the frequency in Hz of the note to sing (Rachel can hit a note
// within 1hz, which is approximatly 10 cents around A below Middle C)
int RachelNoteToFreq(int note)
{

	// Right now we are locked to equal temperment

	//note: midi note number (69 = A = 440Hz)
	//output: frequency in hertz
	switch (note)
		{
		case 57: 	// A below middle C
			return 220;
		case 58:
			return 233;
		case 59:
			return 247;
		case 60:	// Middle C
			return 262;
		case 61:
			return 277;
		case 62:
			return 294;
		case 63:
			return 311;
		case 64: 
			return 330;
		case 65:
			return 349;
		case 66:
			return 370;
		case 67:
			return 392;
		case 68:
			return 415;
		case 69:       // A above middle C
			return 440;
		case 70:
			return 466;
		case 71:
			return 494;
		case 72:
			return 523;
		case 73:
			return 554;
		case 74:
			return 587;
		case 75:
			return 622;
		case 76:
			return 659;
		case 77:
			return 698;
		case 78:
			return 740;
		case 79:
			return 784;
		case 80:
			return 831;
		case 81:
			return 880;
		}

	// Make the note sour if it is out of bounds
	return 60;

}

//Play a specified note for a specified time
void RachelPlayNote(int note, int duration)
{
	// note: midi note number (69 = A = 440hz)
	// duration: duration of sound in ms
	SndCommandType snd;

	snd.cmd = sndCmdFreqDurationAmp;
	snd.param1 = RachelNoteToFreq(note); // frequency
	snd.param2 = duration; // duration (in ms)
	snd.param3 = RachelPrefVolume; // volume
	SndDoCmd(0, &snd, 0);

}

// Given a first note and an interval, play the interval
void RachelPlayInterval(int play_note, int play_interval)
{

	RachelNotePair notes;

	notes = RachelIntervalInRange(play_note, play_interval);
	RachelPlayNote(notes.note1,RachelConvertDuration(RachelPrefDuration));
	RachelPlayNote(notes.note2,RachelConvertDuration(RachelPrefDuration));

}

// Clear out the information message box
void RachelClearInform()
{

	RectanglePtr erase_box;

	// They had better keep RectangleType the structure RectanglrPtr
        // points to
	if((erase_box = MemPtrNew(sizeof(RectangleType))) == NULL)
		{
		// Is there a better way to handle memory starvation?
		FrmAlert(RachelOutOfMemory);
		return;
		}

	// Initialize a rectangle the size of the text field we need to erase
	RctSetRectangle(erase_box,RACHEL_MESSAGE_BOX_X, RACHEL_MESSAGE_BOX_Y,
                        RACHEL_MESSAGE_BOX_WIDTH, RACHEL_MESSAGE_BOX_HEIGHT);

	// And erase the text field	
	WinEraseRectangle(erase_box,0);

	// Now, free the memory we allocated above
	MemPtrFree(erase_box);

}

// Tell the user they guessed the wrong interval
void RachelInformWrongInterval()
{

	// Clear any visible messages
	RachelClearInform();

	// Put in "Wrong Interval" Message
	WinSetUnderlineMode(noUnderline);
	FntSetFont(boldFont);
	WinDrawChars(RACHEL_WRONG_INTERVAL_MESSAGE,RACHEL_WRONG_INTERVAL_MSGL,
		     RACHEL_MESSAGE_BOX_X,RACHEL_MESSAGE_BOX_Y);

	// Put up alert (if the user wants an alert on a wrong message)
	if(RachelPrefAlertOn == RACHEL_ALERTON_WRONG ||
           RachelPrefAlertOn == RACHEL_ALERTON_BOTH)
		FrmAlert(RachelWrongInterval);

}

// Tell the user they determined the correct interval
void RachelInformRightInterval()
{

	// Clear any visible messages
	RachelClearInform();

	// Put in "Right Interval" Message
	WinSetUnderlineMode(noUnderline);
	FntSetFont(boldFont);
	WinDrawChars(RACHEL_RIGHT_INTERVAL_MESSAGE,RACHEL_RIGHT_INTERVAL_MSGL,
		     RACHEL_MESSAGE_BOX_X,RACHEL_MESSAGE_BOX_Y);

	// Put up an alert (if applicable)
	if(RachelPrefAlertOn == RACHEL_ALERTON_RIGHT || 
           RachelPrefAlertOn == RACHEL_ALERTON_BOTH)
		FrmAlert(RachelRightInterval);

}

// When the "random interval" button is pressed, we pop up a dialog if
// the interval has not been guessed yet, and then select and play a new 
// interval
void RachelNewInterval()
{
	
	// If they have not guessed the last interval, confirm that
	// they want to play a new interval
	if(RachelGlobalGuessed == false)
		if(FrmAlert(RachelNotGuessedNew) == RachelNotGuessedNewNo)
			return;

	// We did not guess this new interval yet
	RachelGlobalGuessed = false;
	// Clear any message
	RachelClearInform();

	// Determine the note and interval
	RachelPref1stNote  = RachelRandomNote();
	RachelPrefInterval = RachelRandomInterval();

	// We also select the random values for note duration and direction
	// We do it here because we want "Play interval again" to play the
	// same random direction and duration over and over.
	RachelPrefRndDrtn  = RachelRandomDuration();
	RachelPrefRndDir   = RachelRandomDirection();

	// Play the interval
	RachelPlayInterval(RachelPref1stNote, RachelPrefInterval);

}

// When the "Name notes in interval" button is pressed, check to see if they
// have guessed the interval yet (alert them if they haven't), and then show 
// the notes
void RachelNameNotesPressed()
{

	// If they have not guessed the last interval, confirm that
	// they want to name the notes in the interval
	if(RachelGlobalGuessed == false)
		if(FrmAlert(RachelNotGuessedName) == RachelNotGuessedNameNo)
			return;

	RachelNameNotesInInterval(RachelPref1stNote, RachelPrefInterval);
}

// When the "play current guess" button is pressed, check to see if they
// have guessed the interval yet (alert them if they haven't), and then play
// the interval in question
void RachelPlaySelPressed()
{

	// If they have not guessed the last interval, confirm that
	// they want to play the selected interval
	if(RachelGlobalGuessed == false)
		if(FrmAlert(RachelNotGuessedPlay) == RachelNotGuessedPlayNo)
			return;

	RachelPlayInterval(RachelPref1stNote, RachelPrefGuess);

}

// When the "play interval again" button is pressed, we play the interval
// in question again, or bug them if there isn't an interval to play
void RachelRepeatInterval()
{
	if(RachelPref1stNote > 0 && RachelPrefInterval >= 0)
		RachelPlayInterval(RachelPref1stNote, RachelPrefInterval);
	else
		FrmAlert(RachelNoPlayInterval);

}

// When they select an interval on the right, see if the guess is correct
// or not
void RachelCheckGuess(int selection)
{

	// We keep track of the interval guessed
	RachelPrefGuess = selection;
	// Did they make the right guess
	if(RachelPrefGuess == RachelPrefInterval)
		{
		RachelGlobalGuessed = true;
		RachelInformRightInterval();
		}
	else if(RachelPrefInterval >= 0)
		RachelInformWrongInterval();
	else
		FrmAlert(RachelNoInterval);

}

// When they select a new volume on the volume popup, we change the
// volume in the preferences
void RachelChangeVolume(int selection)
{

	switch(selection)
	{
	case 0:
		RachelPrefVolume = RACHEL_VOLUME_LOW;
		break;
	case 1:
		RachelPrefVolume = RACHEL_VOLUME_MID;
		break;
	case 2:		
		RachelPrefVolume = RACHEL_VOLUME_HIGH;
		break;
	default:
		RachelPrefVolume = RACHEL_VOLUME_MID;
	}

}

// When they select a duration in the duration popup, change the
// preference for the duration
void RachelChangeDuration(int selection)
{

	if(selection >= RACHEL_DURATION_SHORT && 
           selection <= RACHEL_DURATION_MAX)
		RachelPrefDuration = selection;
	else
		RachelPrefDuration = RACHEL_DURATION_MED;

}

// When they select a direction in the direction popup,
// change the preference for the duration
void RachelChangeDirection(int selection)
{

	if(selection >= RACHEL_DIRECTION_UP &&
	   selection <= RACHEL_DIRECTION_MAX)
		RachelPrefDirection = selection;
	else
		RachelPrefDuration = RACHEL_DIRECTION_UP;

}

// When they select when to show an alert (Wrong Interval, Right interval, 
// both), change the preference for the Alert on
void RachelChangeAlertOn(int selection)
{

	if(selection >= RACHEL_ALERTON_RIGHT &&
	   selection <= RACHEL_ALERTON_MAX)
		RachelPrefAlertOn = selection;
	else
		RachelPrefAlertOn = RACHEL_ALERTON_RIGHT;

	// Change the alert on the currently active form
	RachelChangeAlertPopup(NULL);

}

// When they ask for help, give them a series of help screens
void RachelDoHelp()
{
int onpage = 1;  // The help page we are on (0 to exit)
Word status = -1;  // The button pressed at the bottom of help (Next, Previous,
                   // or Done)

while(onpage > 0)	
	{
	// Show the help screen and find out if they pushed
	// Next, Previous, or Done
	switch(onpage) 
		{
		case 1:
			status = FrmAlert(RachelHelpAlert);
			break;
		case 2:
			status = FrmAlert(RachelHelp2Alert);
			break;
		case 3:
			status = FrmAlert(RachelHelp3Alert);
			break;
		case 4:
			status = FrmAlert(RachelHelp4Alert);
			break;
		default: // This should never happen, but just in case
			onpage = 0;
		}

	// If they pushed "Next", go to the next page
	if((onpage == 1 && status == RachelHelpNext) ||
           (onpage == 2 && status == RachelHelp2Next) ||
           (onpage == 3 && status == RachelHelp3Next))
		onpage++;
	else  // Otherwise, they clicked "Done"
		onpage = 0;

	}

}

// Callback to handle all the events on Rachel's single form
static Boolean RachelHandleEvent(EventPtr event)
{
	Boolean		handled;

#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif
	handled = false;

	switch (event->eType)
		{

		case lstSelectEvent: // List was selected
			// Have they selected an interval on the right?
			if(event->data.lstSelect.listID == RachelIntervalList)
			{
			RachelCheckGuess(event->data.lstSelect.selection);
			handled = true;
			}
		break;

		case popSelectEvent:  // Popup list was selected
			switch(event->data.popSelect.listID)
			{

			// Did we ask to change the volume?
			case RachelVolumePopup:
				RachelChangeVolume(event->
                                                   data.popSelect.selection);
				break;

			// Did we ask to change the duration?
			case RachelDurationPopup:
				RachelChangeDuration(event->
						     data.popSelect.selection);
				break;

			// Did we ask to change the direction?
			case RachelDirectionPopup:
				RachelChangeDirection(event->
						     data.popSelect.selection);
				break;

			// Did we ask to change when we make an alert
			case RachelAlertPopup:
				RachelChangeAlertOn(event->
						    data.popSelect.selection);
				handled = true;
				break;

			}
			// We don't make handled true because we want
                        // to perform the default action of selecting the
                        // list choice in question
			break;

   		case ctlSelectEvent:  // Button was pressed and released.
		switch(event->data.ctlSelect.controlID)
			{	
			// Make a random interval for the user to guess,
			// and sing the interval in question
			case RachelPlayRandomButton:
				RachelNewInterval();
				handled = true;
				break;	
			// Play the interval to guess again
			case RachelPlayAgainButton:
				RachelRepeatInterval();
				handled = true;
				break;
			// Disabled feature: play the selected interval
			case RachelPlaySelectedButton:
				RachelPlaySelPressed();
				handled = true;
				break;
			// Tell them the notes played in the interval
			// to guess
			case RachelNameNotesButton:	
				RachelNameNotesPressed();
				handled = true;
				break;
			}	
		break;

		// This is run at startup time
  		case frmOpenEvent:	
			// Restore the list selections from the preferences
			RachelRestorePrefs();
			// Physically draw the form
			FrmDrawForm(FrmGetActiveForm());
			handled = true;
			break;
		
		// Has the user selected something on the menu on top?	
		case menuEvent:		
			if (event->data.menu.itemID == RachelMenuAbout)
   				FrmAlert(RachelAboutAlert);
			else
				RachelDoHelp();
			handled = true;
			break;

		}
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
	return(handled);
}


static Boolean RachelHandleEventWrapper(EventPtr event)
{
	FormPtr	frm;
	Int		formId;
	Boolean	handled = false;

	if (event->eType == frmLoadEvent)
		{
		// Load the form resource specified in the event then activate the form.
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);

		// Set the event handler for the form.  The handler of the currently 
		// active form is called by FrmDispatchEvent each time it receives an event.
		switch (formId)
			{
			case RachelForm:
				FrmSetEventHandler(frm, RachelHandleEvent);
				break;
			}
		handled = true;
		}
	
	return handled;
}

static void RachelEventLoop(void)
{
	EventType	event;
	Word			error;
	
	do
		{
		EvtGetEvent(&event, evtWaitForever);
		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! RachelHandleEventWrapper(&event))
					FrmDispatchEvent(&event);
		}
	while (event.eType != appStopEvent);
}


DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
	Err err = 0;		
	
	if (cmd == sysAppLaunchCmdNormalLaunch)
		{
		if ((err = RachelStartApplication()) == 0)
		{
			RachelEventLoop();
			
			RachelStopApplication();
			}
		}

	// Close all forms to stop memory leaks

	return err;
}
